/*!\file animeManager.h
 *
 * \brief Bibliothèque de gestion des écrans d'animation et de mises à
 * jour en fonction du son.
 * \author Farès BELHADJ, amsi@ai.univ-paris8.fr
 * \date April 29 2014
 */
#ifndef _ANIME_MANAGER_H

#define _ANIME_MANAGER_H

#include <gl4du.h>
#include <primitives.h>

#ifdef __cplusplus
extern "C" {
#endif

  typedef struct anime_t anime_t;
  
  enum anime_state_t {
    S_INIT,
    S_UPDATE_WITH_AUDIO,
    S_DRAW,
    S_FREE
  };
  
  struct anime_t {
    Uint32 time;
    void (* first)      (int state);
    void (* last)       (int state);
    void (* transition) (void (* anim0)(int), void (* anim1)(int), Uint32 fullTime, Uint32 elapsedTime, int state);
  };
  
  extern void   amInit(anime_t * animations);
  extern void   amFree(anime_t * animations);
  extern void   amStartingSignal(void);
  extern Uint32 amGetTicks(void);
  extern void   amDraw(anime_t * animations);
  extern void   amUpdateWithAudio(anime_t * animations);

#ifdef __cplusplus
}
#endif

#endif
